#include "cylinder.h"

// constructeur par dfaut
Cylinder::Cylinder() {}



// constructeur par copie
Cylinder:: Cylinder(const Cylinder &cyl)
{
  nbSections = cyl.nbSections;
  points = cyl.points;
  facets = cyl.facets;
}



// destructeur
Cylinder::~Cylinder()
{  
  clearVector(points);
  facets.clear();
  clearVector(axis);
  clearVector(profile);
  clearVector(section);
}



// constructeur pour l'extrusion simple
Cylinder::Cylinder(const vector<Primitive *> &primAxis,
		   const vector<Primitive *> &primSection)
{
  // on rcupre tous les points d'chantillons de toutes les primitives de l'axe
  for(unsigned int i = 0; i < primAxis.size() ; i++)
    for(unsigned int j = 0  ; j < primAxis[i]->samplesPointsList.size() ; j++)
      axis.push_back(primAxis[i]->samplesPointsList[j]);
  
  // idem pour la section
  for(unsigned int i = 0; i < primSection.size() ; i++)
    for(unsigned int j = 0  ; j < primSection[i]->samplesPointsList.size() ; j++)
      section.push_back(primSection[i]->samplesPointsList[j]);

  // le nombre de sections est le nombre d'chantillons dans l'axe
  nbSections = axis.size();
}



// constructeur pour l'extrusion profil
Cylinder::Cylinder( const vector<Primitive *> &primAxis,
		    const vector<Primitive *> &primSection,
		    const vector<Primitive *> &primProfile)
{
  // on rcupre tous les points d'chantillons de toutes les primitives de l'axe
  for(unsigned int i = 0; i < primAxis.size() ; i++)
    for(unsigned int j = 0  ; j < primAxis[i]->samplesPointsList.size() ; j++)
      axis.push_back(primAxis[i]->samplesPointsList[j]);

  // idem pour la section
  for(unsigned int i = 0; i < primSection.size() ; i++)
    for(unsigned int j = 0  ; j < primSection[i]->samplesPointsList.size() ; j++)
      section.push_back(primSection[i]->samplesPointsList[j]);

  // idem pour le profil
  for(unsigned int i = 0; i < primProfile.size() ; i++)
    for(unsigned int j = 0  ; j < primProfile[i]->samplesPointsList.size() ; j++)
      profile.push_back(primProfile[i]->samplesPointsList[j]);

  // le nombre de sections est le nombre d'chantillons dans l'axe
  nbSections = axis.size();
}



// effacer les vecteurs
void
Cylinder::clearVectors()
{
  clearVector(points);
  facets.clear();
}



// approximation de la tangente :
// la tangente est approxime comme le vecteur entre 2 points chantillons successifs de l'axe
Vector3D *
Cylinder::approximationTangent(Point3D *pt1, Point3D *pt2)
{
  Vector3D tangent = Vector3D(*pt1,*pt2);
  return new Vector3D(tangent.normalize());
}



// calcul des tangentes
vector<Vector3D *>
Cylinder::calculateTangents()
{
  vector<Vector3D *> tangents;
  Point3D *pt1;
  Point3D *pt2;

  //calcul des vecteurs tangents aux sections
  for(int i = 0; i < nbSections - 1; i++)
    {
      pt1 = axis[i];
      pt2 = axis[i+1];
      
      //on ajoute le point a la liste des points origine
      Vector3D * tangent = approximationTangent(pt1,pt2);
      if(tangent != NULL) 
	tangents.push_back(tangent);
    }
  
  tangents.push_back(tangents[nbSections - 2]);
  
  return tangents; 
}



// calcul des points pour l'extrusion simple
void
Cylinder::calculatePoints(vector<Point3D *> &points, const vector<Vector3D *> &tangents)
{
  // caractrisation du repre de Fresnet
  Vector3D normal;                  // normale au point d'chantillon
  Vector3D normalPrev;              // normale prcdente
  Vector3D binormal(0.0,0.0,-1.0);  // binormale (plan Oxy --> binormale constante)

  // nombre de points d'chantillons dans la section
  int rangeOfSamplesSection = section.size();

  // angle de torsion
  double angle = 0.0;
  
  for(int i = 0; i < nbSections; i++)
    {
      // calcul de la normale
      normal = binormal ^ *tangents[i];

      // inversion de la normale si elle est oppose  la prcdente
      if(i != 0)
	if((normal * normalPrev) < 0)
	  normal *= -1;
      
      normalPrev = normal;
      
      // le centre de la section est le point chantillon courant de l'axe
      Point3D sectionCenterPoint = *axis[i];

      // calcul de l'angle si la torsion n'est pas nulle
      if (twistAngle != 0)
	angle += (double) ((twistAngle / nbSections) * M_PI / 180);

      for(int j = 0; j < rangeOfSamplesSection; j++)
	{
	  // rcupration d'un point chantillon de la section
	  Point3D sectionPoint = *section[j];

	  // calcul de la rotation si la torsion n'est pas nulle
	  if (twistAngle != 0)
	    {
	      double tr_x = sectionPoint[0] - (*axis[0])[0];
	      double tr_y = sectionPoint[1] - (*axis[0])[1];

	      sectionPoint[0] = (tr_x * cos(angle) - tr_y * sin(angle)) + (*axis[0])[0];
	      sectionPoint[1] = (tr_x * sin(angle) + tr_y * cos(angle)) + (*axis[0])[1];
	    }
	  
	  // calcul des coordonnes du point
	  double x = sectionCenterPoint[0] + (sectionPoint[0] * normal[0] + sectionPoint[1] * binormal[0]) / SCALE_FACTOR_SIMPLE;
	  double y = sectionCenterPoint[1] + (sectionPoint[0] * normal[1] + sectionPoint[1] * binormal[1]) / SCALE_FACTOR_SIMPLE;
	  double z = sectionCenterPoint[2] + (sectionPoint[0] * normal[2] + sectionPoint[1] * binormal[2]) / SCALE_FACTOR_SIMPLE;
	  
	  // ajout du point  la liste
	  points.push_back(new Point3D(x,y,z));
	}
    }
}




// calcul des facettes
void
Cylinder::calculateFacets(vector<int> &facets)
{
  // nombre de points d'chantillons dans la section
  int rangeOfSamplesSection = section.size();

  // limite variable selon si la section est ferme
  int limit;
  
  if (closedShape)
    limit = rangeOfSamplesSection;
  else
    limit = rangeOfSamplesSection - 1;
  
  
  for(int i = 0 ; i < nbSections -1; i++)
  
    for(int j =0 ; j < limit ; j++)
      {
	//  chaque fois, on rcupre 4 points et on gnre 2 facettes triangulaires
	
	// premier triangle
	facets.push_back(i*rangeOfSamplesSection + j);
	facets.push_back((i+1)*rangeOfSamplesSection + j);
	facets.push_back(i*rangeOfSamplesSection + (j+1) % rangeOfSamplesSection);
	
	// deuxime triangle
	facets.push_back((i+1)*rangeOfSamplesSection + j);
	facets.push_back((i+1)*rangeOfSamplesSection + (j+1) % rangeOfSamplesSection);
	facets.push_back(i*rangeOfSamplesSection + (j+1)%rangeOfSamplesSection);
      } 
}



// calcul des points pour l'extrusion profil
void
Cylinder::calculateProfilePoints(vector<Point3D *> &points,
				 const vector<Vector3D *> &tangents,
				 vector<double> &radius,
				 vector<int> &indexes)
{
  // caractrisation du repre de Fresnet
  Vector3D normale;                  // normale au point d'chantillon
  Vector3D normalePrev;              // normale prcdente
  Vector3D binormale(0.0,0.0,-1.0);  // binormale (plan Oxy --> binormale constante)

  // nombre de points d'chantillons dans la section
  int rangeOfSamplesSection = section.size();

  // vectuer tangente
  Vector3D tangentNorm;

  // angle de torsion
  double angle = 0.0;

  for(unsigned int i = 0; i < indexes.size(); i++)
    {
      
      // vrification de la tangente
      if (!tangents[indexes[i]])
	{ 
	  if (i > 0)
	    tangentNorm = *tangents[indexes[i-1]];
	  else
	    tangentNorm = Vector3D(1,0,0);
	}
      else
	tangentNorm = *tangents[indexes[i]];
      
      // calcul de la normale
      normale = binormale ^ tangentNorm;

      // inversion de la normale si elle est oppose  la prcdente
      if(i != 0)
	if( (normale * normalePrev) < 0 )
	  normale *= -1;
      
      normalePrev = normale;
      
       // le centre de la section est le point chantillon courant de l'axe
      Point3D sectionCenterPoint = *axis[indexes[i]];
      
      // calcul de l'angle si la torsion n'est pas nulle
      if (twistAngle != 0)
	angle += (double) ((twistAngle / nbSections) * M_PI / 180);

      for(int j = 0; j < rangeOfSamplesSection; j++)
	{
	  // rcupration d'un point chantillon de la section
	  Point3D sectionPoint = *section[j];
		  
	   // calcul de la rotation si la torsion n'est pas nulle
	  if (twistAngle != 0)
	    {
	      double tr_x = sectionPoint[0] - (*axis[0])[0];
	      double tr_y = sectionPoint[1] - (*axis[0])[1];

	      sectionPoint[0] = (tr_x * cos(angle) - tr_y * sin(angle)) + (*axis[0])[0];
	      sectionPoint[1] = (tr_x * sin(angle) + tr_y * cos(angle)) + (*axis[0])[1];
	    }

	  // calcul des coordonnes du point
	  double x = sectionCenterPoint[0] + ( (sectionPoint[0]*normale[0] + sectionPoint[1]*binormale[0] ) * radius[i]) / SCALE_FACTOR_PROFILE;
	  double y = sectionCenterPoint[1] + ( (sectionPoint[0]*normale[1] + sectionPoint[1]*binormale[1] ) * radius[i]) / SCALE_FACTOR_PROFILE;
	  double z = sectionCenterPoint[2] + ( (sectionPoint[0]*normale[2] + sectionPoint[1]*binormale[2] ) * radius[i]) / SCALE_FACTOR_PROFILE;

	  // ajout du point  la liste
	  points.push_back(new Point3D(x,y,z));
	}
    }
}



// mthode d'extrusion simple
void
Cylinder::simpleExtrusion(vector<Point3D *> &points,
			  vector<int> &facets)
{    
  // on efface les vecteurs
  clearVectors();
 
  // on calcule les tangentes 
  vector<Vector3D *> tangents = calculateTangents();
 
  // calcul des points
  calculatePoints(points,tangents);
  
  // calcul des facettes
  calculateFacets(facets);
}



// mthode d'extrusion selon le profil
void
Cylinder::profileExtrusion(vector<Point3D *> &points,
			   vector<int> &facets)
{    
  // on efface les vecteurs
  clearVectors();

  // nombre d'echantillons dans le profil
  int rangeOfSamplesProfile = profile.size();
  
  // rapport entre le nombre d'chantillonnage du profil et celui de l'axe
  // si le profil est plus chantillonn que l'axe, il faut sur-chantillonner l'axe 
  int sampleRatio = (int) (rangeOfSamplesProfile / nbSections);

  // points d'chantillonage de l'axe aprs sur-chantillonnage
  vector<Point3D *> overSampleAxis;
  
  // sur-chantillonnage si rapport > 1
  if (sampleRatio > 1)
    {
      sampleRatio++;

      // on rajoute des points d'chantillonnage dans l'axe par interpolation
      for (int i = 0; i < nbSections - 1; i++)
	{
	  double xStep = ( (*axis[i + 1])[0] - (*axis[i])[0] ) / sampleRatio;
	  double yStep = ( (*axis[i + 1])[1] - (*axis[i])[1] ) / sampleRatio;
	  for (int j = 0; j < sampleRatio; j++)
	    {
	      double x = (*axis[i])[0] + j * xStep;
	      double y = (*axis[i])[1] + j * yStep;
	      overSampleAxis.push_back(new Point3D(x,y,0));
	    }
	}
      overSampleAxis.push_back(axis[nbSections - 1]);
      nbSections = overSampleAxis.size();
      
      axis = overSampleAxis;
    }
  
  
  // calcul des tangentes
  vector<Vector3D *> tangents = calculateTangents();
  
  // indices des chantillons de l'axe
  // (il faut grer les rebroussements :
  //  les indices ne sont donc pas forcment dans le mme ordre!)
  int index = 0;
  
  // valeurs min et max des y du profil :
  // sert  "mapper" le profil sur l'axe (tailles diffrentes)
  double y_min = (*profile[0])[1];
  double y_max = y_min;
 
  // recherche de y_min et y_max
  for (int i = 0; i < rangeOfSamplesProfile; i++)
    {
      double y = (*profile[i])[1];
      if (y < y_min)
	{
	  y_min = y;
	  index = i;  // le point de dpart est celui dont le y est minimum et non toujours le 1er de la liste
	}
      if (y > y_max)
	y_max = y;
    }
  
  // distance sur les y
  double diff_y = y_max - y_min;
  
  vector<int> indexes;      // tableau des indices des sections de l'axe  traiter
  vector<double> radius;    // tableau des rayons associs  chaque section
  
  for (int i = 0; i < rangeOfSamplesProfile - 1; i++)
    {
      // gestion des rebroussements
      int sense;
      
      // diffrence sur y entre 2 chantillons successifs du profil : permet de connatre le sens
      float yVariation = (*profile[i + 1])[1] - (*profile[i])[1];
      
      if (yVariation != 0.0)                             // si la variation n'est pas nulle, le sens vaut 1 ou -1
	sense = (int) (yVariation / abs(yVariation));    // selon le signe de cette variation
      else
	sense = 0;                                       // sinon, le sens est nul (2 sections sur le mme point)
      
      double currentRadius = (*profile[i])[0];           // rayon du point courant du profil ie sa valeur x
      double nextRadius = (*profile[i + 1])[0];          // rayon du point suivant (sert pour l'interpolation)
      
      // nombre de sections de l'axe entre 2 points successifs du profil :
      // on "mappe" le profil sur l'axe
      int nbSectionsInter = (int) (nbSections * abs(yVariation) / diff_y);
      
      // pour un meilleur rendu, on force le minimum  2
      if (nbSectionsInter < 2)
	nbSectionsInter = 2;
      
      // calcul du pas du rayon entre 2 sections intermdiaires (interpolation)
      int radiusStep = (int) ( (nextRadius - currentRadius) / (nbSectionsInter - 1));
      
      // calcul des indices et des rayons associs
      for (int j = 0; j < nbSectionsInter - 1; j++)
	{
	  indexes.push_back(index);                          // ajout de l'indice en cours
	  radius.push_back(currentRadius + j * radiusStep);  // calcul du rayon par interpolation
	  
	  // gestion des extrmits :
	  // on ne peut pas reculer si on est au 1er point de l'axe
	  // ni avancer si on est au dernier
	  if ( (index == 0) && (sense == -1) )
	    sense = 0;
	  if ( (index == nbSections - 1) && (sense == 1) )
	    sense = 0;
	  
	  // calcul de l'indice suivant selon les rebroussements
	  index += sense;
	}
    }

  // calcul des points
  calculateProfilePoints(points,tangents,radius,indexes);
  
  // le nombre de sections est chang
  nbSections = radius.size();

  // calcul des facettes
  calculateFacets(facets); 
}

