/** @file cylinder.h */

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                       /
// Projet : Gnrateur de cylindres gnraliss                                          /
//                                                                                       /
// Nom du fichier : cylinder.h                                                          /
//                                                                                       /
// Creation : 20/11/2005                                                                 /
//                                                                                       /
// Description : La classe Cylinder permet de crer des cylindres gnraliss et de leur /
//               appliquer soit la mthode d'extrusion simple soit celle d'extrusion     /
//               profil.
//                                                                                       /
// Auteurs : - Guitteny Fabrice                                                          /
//           - Idiart Baptiste                                                           /
//           - Le Goff Erwan                                                             /
//                                                                                       /
//////////////////////////////////////////////////////////////////////////////////////////

#ifndef CYLINDER_H
#define CYLINDER_H

#include <vector>
#include <math.h>

#include "../primitives/point.h"
#include "../primitives/vector3D.h"
#include "../primitives/primitive.h"
#include "../interface/utilities.h"

using namespace std;

/** facteur de rtrcissement pour l'extrusion simple */
const int SCALE_FACTOR_SIMPLE = 4;

/** facteur de rtrcissement pour l'extrusion simple */
const int SCALE_FACTOR_PROFILE = 20;


/** La classe Cylinder permet de crer des cylindres gnraliss et de leur
 *  appliquer soit la mthode d'extrusion simple soit celle d'extrusion
 *  profil.
 *
 * @author Guitteny Fabrice
 * @author Idiart Baptiste
 * @author Le Goff Erwan
 */
class Cylinder
{
  private :
    
    /** nombre de sections du cylindre */
    int nbSections;
  
  /** ensemble des points gnrs */
  vector<Point3D *> points;
  
  /** ensemble des facettes gnres */
  vector<int> facets;
  
  /** points chantillons de l'axe */
  vector<Point3D *> axis;

  /** points chantillons de la section */
  vector<Point3D *> section;

  /** points chantillons du profil */
  vector<Point3D *> profile;

  /** savoir si la section est ferme */
  bool closedShape;

  /** angle de torsion */
  int twistAngle;
  
  /** effacer les vecteurs */
  void clearVectors();

  /** calcul des tangentes en chaque point chantillon de l'axe
   *  @return vecteur tangent  l'axe
   */
  vector<Vector3D *> calculateTangents();

  /** calcul des points pour l'extrusion simple
   *  @param points ensemble des points
   *  @param tangents ensemble des tangentes associes
   */
  void calculatePoints(vector<Point3D *> &points,
		       const vector<Vector3D *> &tangents);

  /** calcul des points pour l'extrusion selon le profil
   *  @param points ensemble des points
   *  @param tangents ensemble des tangentes associes
   *  @param radius rayons associs
   *  @param indexes indices des points chantillons de l'axe
   */
  void calculateProfilePoints(vector<Point3D *> &points,
			      const vector<Vector3D *> &tangents,
			      vector<double> &radius,
			      vector<int> &indexes);

  /** calculs des facettes
   *  @param facets ensemble des facettes
   */ 
  void calculateFacets(vector<int> &facets);
  
  public :
    
    /** constructeur par dfaut */
  Cylinder();

  /** constructeur par copie
   *  @param cyl cylindre gnralis
   */
  Cylinder(const Cylinder &cyl);
  
  /** constructeur pour l'extrusion simple
   *  @param primAxis ensemble des primitives de l'axe
   *  @param primSection ensemble des primitives de la section
   */
  Cylinder( const vector<Primitive *> &primAxis,
	    const vector<Primitive *> &primSection);
  
  /** constructeur pour l'extrusion selon le profil
   *  @param primAxis ensemble des primitives de l'axe
   *  @param primSection ensemble des primitives de la section
   *  @param primProfile ensemble des primitives du profil
   */
  Cylinder( const vector<Primitive *> &primAxis,
	    const vector<Primitive *> &primSection,
	    const vector<Primitive *> &primProfile);
  
  // destructeur
  ~Cylinder();
  
  /** approximation de la tangente
   *  @param pt1 point1
   *  @param pt2 point2
   *  @return vecteur tangente
   */
  Vector3D * approximationTangent(Point3D *pt1, Point3D *pt2);
  
  /** mthode d'extrusion simple
   *  @param points ensemble des points
   *  @param facets ensemble des facettes
   */
  void simpleExtrusion(vector<Point3D *> &points,
		       vector<int> &facets);

  /** mthode d'extrusion selon le profil
   *  @param points ensemble des points
   *  @param facets ensemble des facettes
   */
  void profileExtrusion(vector<Point3D *> &points,
		       vector<int> &facets);

  /** modificateur : si la section est ferme
   *  @param closed vrai si la section est ferme
   */
  inline void setClosedSection(bool closed) {closedShape = closed;}

  /** modificateur : valeur de l'angle de torsion
   *  @param angle nouvelle valeur de l'angle
   */ 
  inline void setTwist(int angle) {twistAngle = angle;}

};

#endif

