#include "window2D.h"


// ****************
// * constructeur *
// ****************

Window2D::Window2D( QWidget *parent, const char *name ) : QWidget( parent, name )
{
  // initialisation de la fentre
  setBackgroundColor( white ); 
  setCaption(name);
  setBaseSize( QSize(WIDTH,HEIGHT) );
  
  // initialisation des paramtres de la fentre
  nbPoints = 0;
  nbPrimitives = 0;
  down = FALSE;
  move = FALSE;
  closedShape = FALSE;
  
  origin = Point3D(width() / 2, height() / 2, 0);
  
  indexCurrentPoint = 0;
  toBezier = FALSE;

  sample = NORMAL_SAMPLE;
  
  // type de prinitive par dfaut --> SEGMENT
  typePrimitive = SEGMENT;
  
  modification = FALSE;
}

// ***************
// * destructeur *
// ***************

Window2D::~Window2D()
{
  clearVectors();
}


// **************************************
// * effacer les vecteurs de la fentre *
// **************************************

void
Window2D::clearVectors()
{
  primitivesList.clear();
  pointsList.clear();
  lastPointsList.clear();
}



// ************************
// * dessin de la fentre *
// ************************

void
Window2D::paintEvent( QPaintEvent * )
{
  
  // cration et initialisation d'un painter
  QPainter paint( this );
  QPen pen(black,5);
  paint.setPen(pen);
  paint.setWindow(0,0,width(),height());
  paint.setViewport(0,0,width(),height());
  paint.setBrushOrigin(width()/2,height()/2);

  // changement de repre
  if (origin[0] != width()/2 || origin[1] != height()/2)
    {
      changeCoordinateSystemWindow();
      origin[0] = width()/2;
      origin[1] = height()/2;
      changeCoordinateSystemReverseWindow();
    }
 
  // dessin des axes
  int x0 = paint.brushOrigin().x();
  int y0 = paint.brushOrigin().y();
  Point3D * p1 = new Point3D(x0 + width()/2, y0, 0);
  Point3D * p2 = new Point3D(x0,  y0 + width()/2, 0);
  Point3D * p3 = new Point3D(x0 - width()/2, y0, 0);
  Point3D * p4 = new Point3D(x0, y0 - width()/2, 0);
  Segment * axisX = new Segment(p1,p3);
  Segment * axisY = new Segment(p2,p4);
  axisX->display(paint);
  axisY->display(paint);
  
  delete axisX;
  delete axisY;
  
  // dessin de tous les points dfinis par l'utilisateur
  for (unsigned int i = 0; i < pointsList.size(); i++)
    // on dessine un rectangle pour les points afin qu'ils soient bien visibles
    paint.drawRect(pointsList[i].x()-2,pointsList[i].y()-2,4,4);

  // si l'utilisateur a cr des primitives, on les dessine
  if (nbPrimitives > 0)
    for (int i = 0; i < nbPrimitives; i++)
      primitivesList[i]->display(paint);  

  modification = TRUE;  
}



// **********************************
// * vnement souris : relchement *
// **********************************

void
Window2D::mouseReleaseEvent( QMouseEvent * )
{
  down = FALSE;              // la souris est relche
  move = FALSE;              // l'utilisateur ne peut plus bouger de primitives
  update();                  // appel  la fonction de dessin
}



// ********************************
// * vnement souris : mouvement *
// ********************************

void
Window2D::mouseMoveEvent( QMouseEvent * e)
{
  // si le bouton droit de la souris est press,
  // l'utilisateur peut bouger les primitives
  
  if (move && down)
    {
      // on rcupre le point initial  bouger
      QPoint refPoint = pointsList[selectedPoint];

      // on rcupre la nouvelle position
      QPoint modifPoint = e->pos();

      // on vrifie si cette nouvelle position est "proche" d'un autre point dj dfini
      int nearTest = nearPoint(e->pos(), pointsList);

      if (nearTest == selectedPoint || nearTest == -1)
	{
	  // si on n'est pas proche d'un autre point,
	  // on modifie les coordonnes du point
	  pointsList[selectedPoint] = modifPoint;
	  
	  // et on modifie toutes les primitives bases sur ce point
	  for(unsigned int i = 0; i < primitivesList.size(); i++)
	    primitivesList[i]->setPrimitive(refPoint,modifPoint);

	  currentPoint = pointsList[indexCurrentPoint];

	  lastPointsList[0] = qPointToPoint3D(currentPoint);
	}

      // appel  la fonction de dessin
      update();  
    }     
}


// *******************************
// * vnement souris : pression *
// *******************************

void
Window2D::mousePressEvent( QMouseEvent *e )
{
  // la souris est presse
  down = TRUE;

  // bouton de gauche --> dessin d'un point
  if (e->button() == LeftButton)
    {
      // cas particulier : l'utilisateur passe du dessin d'une bspline  une autre primitive
      // problme : si le nouveau type de primitives est une bezier, cette dernire n'est
      //            pas dessine tant qu'il n'a pas dfini 4 points et donc ce test est vrai
      //            tant que les 4 points ne sont pas dfinis
      if ( (nbPrimitives > 0) && (primitivesList[nbPrimitives-1]->type == BSPLINE) && (typePrimitive != BSPLINE) )
	{
	  if (typePrimitive != BEZIER)
	    {
	      if (toBezier == FALSE)
		{
		  lastPointsList.clear();
		  lastPointsList.push_back(qPointToPoint3D(currentPoint));
		}
	    }
	  else
	    {
	      if (toBezier == FALSE)
		{
		  lastPointsList.clear();
		  lastPointsList.push_back(qPointToPoint3D(currentPoint));
		  toBezier = TRUE;
		}
	    }  
	}

      int nearTest = nearPoint(e->pos(), pointsList);
      
      // on vrifie si le point dsign par l'utilisateur est dj dfini
      
      if (nearTest == -1)                                      // si ce n'est pas le cas, on ajoute ce point
	{                                                      // et on met  jour le point courant
	  indexCurrentPoint = pointsList.size();
	  pointsList.push_back(e->pos());
	  currentPoint = e->pos();
	}
      else                                                     // sinon, le point courant est le point dj dfini
	{
	  indexCurrentPoint = nearTest;
	  closedShape = TRUE;
	  currentPoint = pointsList[indexCurrentPoint];
	}
      
      lastPointsList.push_back(qPointToPoint3D(currentPoint)); // on cre le point3D associ au point courant et on l'ajoute  la liste
      
      nbPoints = pointsList.size();                            // on met  jour le nombre de points dfinis
      
      // on cre les primitives selon le type selectionn
      
      switch (typePrimitive)
	{
	case SEGMENT :
	  if (lastPointsList.size() >= 2)			           // il faut au moins 2 points pour dessiner un segment
	    {
	      for (unsigned int i = 0; i < lastPointsList.size() - 1; i++)
		{
		  Segment *seg = new Segment();                           // on cre un segment  partir des 2 derniers points dfinis par l'utilisateur
		  seg->pointsList.push_back(lastPointsList[i]);         
		  seg->pointsList.push_back(lastPointsList[i + 1]);
		  
		  primitivesList.push_back(seg);                          // on ajoute le segment cr  la liste de primitives de la fentre
		  nbPrimitives++;                                         // on incrmente le nombre de primitives
		}
	      lastPointsList.clear();
	      lastPointsList.push_back(qPointToPoint3D(currentPoint));
	    }
	  break;
	  
	case CIRCLE :
	  
	  if (lastPointsList.size() >= 2)                                 // il faut au moins 2 points pour dessiner un cercle
	    {                
	      for (unsigned int i = 0; i < lastPointsList.size() - 1; i++)
		{                 
	          Circle *cir = new Circle();                             // on cre un cercle  partir des 2 derniers points dfinis par l'utilisateur
	          cir->pointsList.push_back(lastPointsList[i + 1]);
	          cir->pointsList.push_back(lastPointsList[i]);
		  cir->calculateRadius();                                 // on calcule le rayon
		  
		  primitivesList.push_back(cir);                          // on ajoute le cercle cr  la liste de primitives de la fentre
		  nbPrimitives++;                                         // on incrmente le nombre de primitives
		}
	      
	      // on repart du point dfinissant le rayon du cercle et non de son centre
	      Point3D * lastButOnePoint = lastPointsList[lastPointsList.size() - 2];  
	      lastPointsList.clear();
	      lastPointsList.push_back(lastButOnePoint);	      						
	    }
	  break;
	  
	case BEZIER :
	  if (lastPointsList.size() >= 4)                                // il faut au moins 4 points pour dessiner une bzier
	    {
	      for (unsigned int i = 0; i < lastPointsList.size() - 1; i += 4)
		{
		  Bezier * bez = new Bezier();                           // on cre une courbe de Bzier  partir des 4 derniers points dfinis par l'utilisateur
		  bez->pointsList.push_back(lastPointsList[i]);
		  bez->pointsList.push_back(lastPointsList[i + 1]);
		  bez->pointsList.push_back(lastPointsList[i + 2]);
		  bez->pointsList.push_back(lastPointsList[i + 3]);
		  
		  primitivesList.push_back(bez);                        // on ajoute la courbe de Bzier cre  la liste de primitives de la fentre
		  nbPrimitives++;                                       // on incrmente le nombre de primitives
		}
	      lastPointsList.clear();
	      lastPointsList.push_back(qPointToPoint3D(currentPoint)); 
	      
	      toBezier = FALSE;
	    }
	  break;
	  
	case BSPLINE :
	  if (lastPointsList.size() >= 4)                              // il faut au moins 4 points pour dessiner une bspline
	    {
	      Bspline * bspl = new Bspline();   
	      
	      // si la dernire primitive dessine n'est pas une bspline,
	      if ( (nbPrimitives == 0) || ( (nbPrimitives > 0) && (primitivesList[nbPrimitives-1]->type != BSPLINE) ) )   
		{         
		  for (unsigned int i = 0; i < lastPointsList.size(); i++)
		    bspl->pointsList.push_back(lastPointsList[i]);        
		                                           
		  primitivesList.push_back(bspl);                                     // on ajoute la bspline  la liste des primitives
		  nbPrimitives++;                                                     // on incrmente le nombre de primitives
		}
              else                                                                    // sinon, on met  jour la dernire primitive de l'ensemble
		{
		  int max = primitivesList[nbPrimitives-1]->pointsList.size();
		  
		  for (int i = 0; i < max; i++)
		    {
		      Point3D * point = primitivesList[nbPrimitives-1]->pointsList[i];
		      bspl->pointsList.push_back(point);
		    }
		    
		  if (toBezier)
		    {
		      for (unsigned int i = 1; i < lastPointsList.size(); i++)
	                {
		          bspl->pointsList.push_back(lastPointsList[i]);
	                }
		      toBezier = FALSE;
		    }
		  else   
		    bspl->pointsList.push_back(qPointToPoint3D(currentPoint)); 
		    
		  primitivesList.pop_back();                                          // on remplace l'ancienne bspline par la nouvelle
		  primitivesList.push_back(bspl);            
		}
	    }
	  break; 
	   
	default :
	  break;
	}
      
    }
  else
    {
      if (e->button() == MidButton)                                // si l'utilisateur clique sur le bouton droit de la souris --> on ferme la forme
	{                                                          
	  if (nbPoints > 2)                                        // il faut au moins 3 points pour pouvoir fermer la forme
	    {
	      Segment *seg = new Segment();                        // on cre le segment entre le point en cours et le 1er point de la liste 
	      seg->pointsList.push_back(qPointToPoint3D(currentPoint));
	      Point3D * refPoint = qPointToPoint3D(pointsList[0]);
	      seg->pointsList.push_back(refPoint);	      
	      lastPointsList.clear();
	      lastPointsList.push_back(refPoint);
	      primitivesList.push_back(seg);                        // on ajoute ce segment  la liste de primitives
	      nbPrimitives++;  
	      closedShape = true;
	    }
	  else
	    {
	      // sinon, on envoie un message d'erreur
	      QMessageBox *mbox = new QMessageBox(QString("ERREUR"),QString("Il faut au moins 3 points pour fermer la forme !"),
						  QMessageBox::Critical,QMessageBox::Ok,QMessageBox::NoButton,
						  QMessageBox::NoButton,this);
	      mbox->show();
	    }
	}
      else
	{
	  // boutton de droite --> mode dplacement de point
	  // on test si l'utilisateur slectionne bien un point de la fentre
	  int nearTest = nearPoint(e->pos(), pointsList);
	  
	  // si c'est le cas et qu'il n'est pas en train de bouger un autre,
	  // on met  jour la valeur du point slectionn
	  if (nearTest != -1 && !move)
	    {
	      move = TRUE;
	      selectedPoint = nearTest;
	    }	  
	}
    } 
}


// *******************************************
// * chargement d'un ensemble de primitives  *
// *******************************************

void 
Window2D::load(vector<Primitive*> &prim)
{
  // rinitialisation des donnes de la fentre
  clearVectors();
  
  // rcupration des primitives du fichier
  nbPrimitives = prim.size();
  primitivesList = prim;

  closedShape = FALSE;
  sample = NORMAL_SAMPLE;

  // cration des points de contrle
  for (int i = 0; i < nbPrimitives; i++)
    {
      for (unsigned int j = 0; j < primitivesList[i]->pointsList.size(); j++)
	{
	  QPoint point = point3DToQPoint(primitivesList[i]->pointsList[j]);
	  int nearTest = nearPoint(point,pointsList);
	  if (nearTest == -1)
	    pointsList.push_back(point);
	  else
	    closedShape = TRUE;
	}
    }

  // on initialise le point en cours au 1er point de la liste de points
  indexCurrentPoint = 0;
  if (nbPrimitives != 0)
    {
      currentPoint = pointsList[indexCurrentPoint];
      lastPointsList.push_back(qPointToPoint3D(currentPoint));
    }
  update();
  toBezier = FALSE;
} 


// ***************************************************
// * effacer la fentre et rinitialiser les donnes *
// ***************************************************

void
Window2D::clear()
{
  clearVectors();
  nbPrimitives = 0;
  indexCurrentPoint = 0;
  toBezier = FALSE;
  closedShape = FALSE;
  update();
}


// ***************************
// * fermeture de la fentre *
// ***************************

void
Window2D::closeEvent(QCloseEvent * ) {}


// ************************
// * changement de repre *
// ************************

void
Window2D::changeCoordinateSystemWindow()
{
  for (int i = 0; i < nbPrimitives; i++)
    {
      for (unsigned int j = 0; j < primitivesList[i]->samplesPointsList.size(); j++)
	{
	  primitivesList[i]->samplesPointsList[j] = new Point3D(changeCoordinateSystem(*(primitivesList[i]->samplesPointsList[j]),origin));
	}

      for (unsigned int j = 0; j < primitivesList[i]->pointsList.size(); j++)
	{
	  primitivesList[i]->pointsList[j] = new Point3D(changeCoordinateSystem(*(primitivesList[i]->pointsList[j]),origin));
	}
    }
  for (unsigned int i = 0; i < pointsList.size(); i++)
    {
      Point3D * point = qPointToPoint3D(pointsList[i]);
      pointsList[i] = point3DToQPoint(new Point3D(changeCoordinateSystem(*point,origin)));
    }
  for (unsigned int i = 0; i < lastPointsList.size(); i++)
    {
      lastPointsList[i] = new Point3D(changeCoordinateSystem(*lastPointsList[i],origin));
    }
}

// ********************************
// * changement de repre inverse *
// ********************************

void
Window2D::changeCoordinateSystemReverseWindow()
{
  for (int i = 0; i < nbPrimitives; i++)
    {
      for (unsigned int j = 0; j < primitivesList[i]->samplesPointsList.size(); j++)
	{
	  primitivesList[i]->samplesPointsList[j] = new Point3D(changeCoordinateSystemReverse(*(primitivesList[i]->samplesPointsList[j]),origin));
	}

      for (unsigned int j = 0; j < primitivesList[i]->pointsList.size(); j++)
	{
	  primitivesList[i]->pointsList[j] = new Point3D(changeCoordinateSystemReverse(*(primitivesList[i]->pointsList[j]),origin));
	}
    }
  for (unsigned int i = 0; i < pointsList.size(); i++)
    {
      Point3D * point = qPointToPoint3D(pointsList[i]);
      pointsList[i] = point3DToQPoint(new Point3D(changeCoordinateSystemReverse(*point,origin)));
    }
  for (unsigned int i = 0; i < lastPointsList.size(); i++)
    {
      lastPointsList[i] = new Point3D(changeCoordinateSystemReverse(*lastPointsList[i],origin));
    }
}
