/** @file window3D.h */

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                       /
// Projet : Gnrateur de cylindres gnraliss                                          /
//                                                                                       /
// Nom du fichier : window3D.h                                                           /
//                                                                                       /
// Creation : 19/11/2005                                                                 /
//                                                                                       /
// Description : La classe Window3D hrite de QglWidget et permet l'affichage en 3d      /
//               du rsultat de l'extrusion de l'import Vrml, de l'import PGN.           /
//                                                                                       /
//                                                                                       /            
// Auteurs : - Guitteny Fabrice                                                          /
//           - Idiart Baptiste                                                           /
//           - Le Goff Erwan                                                             /
//                                                                                       /
//////////////////////////////////////////////////////////////////////////////////////////

#ifndef WINDOW3D_H
#define WINDOW3D_H

#include <qtimer.h>
#include <sstream>
#include <qgl.h>
#include "vrmlparser.h"
#include "../primitives/vector3D.h"

const float TRA_INC = 0.2;//incrment pour les tranlsations sur les 3 axes
const float ZOO_INC = 0.2;//incrment du zoom pour le Scale
const int WIDTHOGL  = 320;//largeur par dfaut de la fentre
const int HEIGHTOGL = 320;//hauteur par dfaut de la fentre


/**
 * La classe Window3D hrite de QGLWidget et permet l'affichage en 3d
 *  du rsultat de l'extrusion de l'import Vrml ou de l'import PGN.
 *
 *  @author Guitteny Fabrice
 *  @author Idiart Baptiste
 *  @author Le Goff Erwan
 */

///////////////////////////////////////////////////////////////////////////////////////////
//  Classe Window3D                                                                      //
///////////////////////////////////////////////////////////////////////////////////////////

class Window3D : public QGLWidget
{
  Q_OBJECT
    private:
  /** Temporisateur utilis pour un rafraichissement frquent de la fenetre
   */
  QTimer *timer;
  /**liste de points
  */
  vector<Point3D*> pointsList;
  /**liste de rfrences vers les points
  */
  vector<int> referencesList;
  /**liste de vecteurs normales
  */
  vector<Vector3D*> normalsList;
  /**nombre de points
  */
  int nbPoints;
  /**nombre de polygones
  */
  int nbPolys;
  /**sauvegarde de l'ex position de la souris
  */
  int oldX, oldY;
  /**angle de rotation
  */
  float angle[3];
  /**facteur de translation
  */
  float tra[3];
  /**zoom (le meme sur les 3 axes)
  */
  float zoom;
  /**bouton press
  */
  int button;
  /** Pour pouvoir observer le modle sous tous ces angles, deux lumieres sont utilises
  *position de la premire lampe
  */
  GLfloat light1Position[4];
  /**position de la deuxime lampe
  */
  GLfloat light2Position[4];
  /**paramtres ambients
  */
  GLfloat ambientLight[4];
  /**paramtres diffusions
  */
  GLfloat diffuseLight[4];
  /**paramtres spculaires
  */
  GLfloat specularLight[4];
  /**numro du modele dans la displaylist
  *  l'utilisation d'un modele prcompil permet un gain en fps important
  */
  GLuint model;


  
 public:

  /** Constructeur de FenetreOgl
   * @param parent pointeur sur le parent (la fentre principale)
   * @param name nom de la fenetre (Visualisation 3D)
   * @param timeInterval temporisateur pour le rafraichissement
   */
  Window3D( QWidget *parent, char *name, int timeInterval = 50);

   /** Destructeur de FenetreOgl
   */
  virtual ~Window3D();
  
   /** chargement d'un modele 3D
   * @param nb_pts nombre de points
   * @param nb_pls nombre de polygnes
   * @param L_pts liste de points
   * @param L_ref liste contenant les rfrences vers les point on 
   *              construit des triangles donc un polygone est dfini
   *              par trois rfrence (L_ref.size() = 3 *nb_pls) 
   * @return variable de controle d'erreur
   */
  int load(const int &nb_pts, const int &nb_pls, const vector<Point3D*> &L_pts, const vector<int> &L_ref);
  
  /** Effacer la fentre Qgl
   */
  void clear();

  /** Dessin des axes du repre 3D
   */
  void drawAxis();
  
  /** export du modele
   * @param nb_pts nombre de points
   * @param nb_pls nombre de polygnes
   * @param L_pts liste de points
   * @param L_ref liste contenant les rfrences vers les point on 
   *              construit des triangles donc un polygone est dfini
   *              par trois rfrence (L_ref.size() = 3 *nb_pls) 
   * @return variable de controle d'erreur
   */
  int exportOGL(int &nb_pts,int &nb_pls, vector<Point3D*> &L_pts, vector<int> &L_ref);
  
  /** 
  * @return true si le nombre" de points ou le nombre de polygones est nuls
  */
  inline bool isEmpty(){return ((nbPoints == 0) || (nbPolys == 0));}

  protected slots:
  
  void timeOutSlot();

 protected:
  
  /** Initialisation de la fentre Qgl
   */
  void initializeGL();
  
  /** Adaptation de la taille de la fentre
  * @param width nouvelle largeur de la fentre
  * @param height nouvelle hauteur de la fentre
  */
  void resizeGL( int width, int height );
   
  /** Fonction d'affichage
  */
  void paintGL();
   
  /** Modlisation  l'aide d'une display list
  */
  GLuint drawModel();

  /** Rotation
  */
  void rotate ();
  /** Translation
  */
  void translate ();

  /** Scale (zoom)
  */
  void scale ();

  /** Dtection de la pression d'un bouton de la souris
  * @param event venement source (coordonnes, bouton)
  */
  void mousePressEvent(QMouseEvent *event);

  /** Dtection de mouvement de la mollette de la souris
  * @param e venement source (coordonnes, bouton)
  */
  void wheelEvent( QWheelEvent * e );
  
  /** Dtection de mouvement de la souris
  * @param event venement source (coordonnes, bouton)
  */
  void mouseMoveEvent(QMouseEvent *event);
    
  /** Dtection de la pression d'une touche clavier
  * @param e touche presse
  */ 
  void keyPressEvent( QKeyEvent *e );

  /** Action a effectue lorsque le timer arrive  terme
  */
  void timeOut();

  /** Action a effectu lors de la fermeture
  * le corps de cette fonction est vide pour interdire sa fermeture
  */
  void closeEvent(QCloseEvent * );


};

#endif

